-------------------------------------------------------------------------------------
-- | Namespace Definitions |---------------------------------------------------------
-------------------------------------------------------------------------------------

-- Retrieve the addon name and the namespace table.
local addon, npcg = ...

local functua = npcg.functua

local my_table = { }

functua.table = my_table

-------------------------------------------------------------------------------------
-- | Type Definitions | -------------------------------------------------------------
-------------------------------------------------------------------------------------

-- type Table a b :: { a, b }
-- A table is a collection of items of keys and values.

-- type Action a :: (a -> ())
-- An action is a function that takes "a" as an argument, and returns void.

-------------------------------------------------------------------------------------
-- | Iteration Functions | ----------------------------------------------------------
-------------------------------------------------------------------------------------

-- forEach :: Action a -> Table _ a -> ()
-- Performs the same action on every value item in a table.
function my_table.forEach(a, xs)
	for _,x in pairs(xs) do
		a(x)
	end
end

-- keys :: Table a _ -> { a }
-- Returns a table of keys from a table.
function my_table.keys(xs)
  local output = { }
  for k,_ in pairs(xs) do
    table.insert(output, k)
  end
  return output
end

-- values :: Table _ a -> { a }
-- Returns a table of values from a table.
function my_table.values(xs)
  local output = { }
  for _,v in pairs(xs) do
    table.insert(output, v)
  end
  return output
end

-- map :: (a -> b) -> Table k a -> Table k b
-- Applies a function to every item in a table and returns the result.
-- Ordering is preserved.
function my_table.map(f, xs)
	local output = { }
	for k,v in pairs(xs) do
		output[k] = f(v)
	end
	return output
end

-- map2 :: (k -> v -> w) -> Table k v -> Table k w
-- Applies a function to every key and value in a table and returns the result.
function my_table.mapKV(f, xs)
	local output = { }
	for k,v in pairs(xs) do
		output[k] = f(k,v)
	end
	return output
end

--[[
-- filter :: (a -> bool) -> Table k a -> table a
-- Returns a table of all elements in a table that satisfy a predicate.
function my_table.filter(p, xs)
	local output = { }
	for _,v in pairs(xs) do
		if p(v) then 
			table.insert(output, v)
		end
	end
  print "table.filter called, check whether this is intended, as it goes Table->List"
	return output
end
]]

-- filterK :: (a -> bool) -> Table a b -> Table a b
-- Returns table of key-value-pairs of which the key satisfies a predicate.
function my_table.filterK(p, xs)
  local output = { }
	for k,v in pairs(xs) do
		if p(k) then 
			output[k] = v
		end
	end
  print "table.filterK called"
	return output
end

-- filterV :: (b -> bool) -> Table a b -> Table a b
-- Returns a table of all key-value-pairs of which the value satisfies a predicate.
function my_table.filterV(p, xs)
  local output = { }
	for k,v in pairs(xs) do
		if p(v) then 
			output[k] = v
		end
	end
	return output
end

-- filterKeysByValue :: (a -> bool) -> { k -> a } -> { k }
-- Returns a table of all keys in a table where the value satisfies a certain predicate.
function my_table.filterKeysByValue(p, xs)
  local output = { }
  for k,v in pairs(xs) do
    if p(v) then 
      table.insert(output, k)
    end
  end
  return output
end

-- foldr :: (a -> b -> b) -> b -> Table _ a -> b
-- Applies an accumulator function on every element of a given table.
function my_table.foldr(op, val, xs)
	for _,v in pairs(xs) do
		val = op(val, v)
	end
	return val
end

-------------------------------------------------------------------------------------
-- | Combination Functions | --------------------------------------------------------
-------------------------------------------------------------------------------------

-- concatKeys :: { _,{ a,_ } } -> { a,_ }
-- concatKeys :: Table _ (Table a _) -> { a, _ }
-- Concatenates a table of tables.
function my_table.concatKeys(xs)
	local output = { }
	for _,v in pairs(xs) do
		for kv,_ in pairs(v) do
			output[kv] = 1
		end
	end
	return functua.table.keys(output)
end

-------------------------------------------------------------------------------------
-- | Cloning Functions | ------------------------------------------------------------
-------------------------------------------------------------------------------------

-- Returns a clone of a table.
function my_table.clone(xs)
	return functua.table.map(functua.id, xs)
end

-------------------------------------------------------------------------------------
-- | Query Functions | --------------------------------------------------------------
-------------------------------------------------------------------------------------

-- any :: (a -> bool) -> Table _ a -> bool
-- Returns whether any of the elements in a table satisfies a predicate.
function my_table.any(p, xs)
	for _,v in pairs(xs) do
		if p(v) then
			return true
		end
	end
	return false
end

-- all :: (a -> bool) -> Table _ a -> bool
-- Returns whether any of the elements in a table satisfies a predicate.
function my_table.all(p, xs)
	return not my_table.any(
		function(x) return not p(x) end, 
		xs)
end

-- contains :: a -> { a } -> bool
-- Returns whether a value is contained in a table.
function my_table.contains(val, xs)
	return my_table.any(
		function(x) return x == val end, 
		xs)
end

-- containsAny :: { _,b } --> { b } --> bool
-- Returns whether a table contains any of the values of a second table.
function my_table.containsAny(xs, values)
  return my_table.any(
    function(x) return my_table.contains(x, xs) end,
    values)
end

-- containsAny :: { _,b } --> { b } --> bool
-- Returns whether a table contains all of the values of a second table.
function my_table.containsAll(xs, values)
  return my_table.all(
    function(x) return my_table.contains(x, xs) end,
    values)
end

-- without :: { a } -> { a } -> { a }
-- Removes all occurances of certain values from a table.
function my_table.without(xs, values)
  return functua.table.filterV(
    function(x) return not functua.table.contains(x, values) end,
    xs)
end

-- findKey :: 
-- Returns
function my_table.findKey(p, xs)
  for k,v in pairs(xs) do
    if p(v) then return k end
  end
  return nil
end

-------------------------------------------------------------------------------------
-- | Counting Functions | -----------------------------------------------------------
-------------------------------------------------------------------------------------

-- count :: { a } -> int
-- Counts the number of elements in a table.
function my_table.count(xs)
	return my_table.foldr(
    function(accum, x) return accum + 1 end, 0, 
    xs)
end

-- sum :: Num a => { a } -> a
-- Computes the sum of the items in the table.
-- Items must be numbers.
function my_table.sum(xs)
  return my_table.foldr(
    function(accum, x) return accum + x end, 0,
    xs)
end

-- sumBy :: Num b => { a } -> (a -> b) -> b
-- Computes the sum of the items in the table, using a function to convert the items into numbers.
function my_table.sumBy(xs, f)
  return my_table.foldr(
    function(accum, x) return accum + f(x) end,
    0,
    xs)
end

-- End of File.